import { logger } from "@libp2p/logger";
import { bytesToHex as hex } from "@noble/hashes/utils";
import { equals } from "uint8arrays";
import { assert } from "@canvas-js/utils";
import { ConflictError } from "@canvas-js/gossiplog/errors";
/**
 * This differs from the sync function exported from @canvas-js/okra in three ways
 * 1) It only yields keys, not entries
 * 2) It assumes that entries are immutable (and throws an error on conflicts)
 * 3) It only yields keys for entries present in `source` and missing in `target`
 */
export class Driver {
    constructor(topic, source, txn) {
        Object.defineProperty(this, "source", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: source
        });
        Object.defineProperty(this, "txn", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: txn
        });
        Object.defineProperty(this, "log", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.log = logger(`canvas:gossiplog:[${topic}]:driver`);
    }
    async *sync() {
        const sourceRoot = await this.source.getRoot();
        assert(sourceRoot.key === null, "invalid source root");
        this.log("source root: level %d, hash %s", sourceRoot.level, hex(sourceRoot.hash));
        const targetRoot = this.txn.getRoot();
        assert(targetRoot.key === null, "invalid target root");
        this.log("target root: level %d, hash %s", targetRoot.level, hex(targetRoot.hash));
        if (sourceRoot.level === 0) {
            return;
        }
        else if (sourceRoot.level === targetRoot.level && equals(sourceRoot.hash, targetRoot.hash)) {
            return;
        }
        yield* this.syncRoot(sourceRoot, targetRoot.level);
    }
    async *syncRoot(sourceNode, targetLevel) {
        if (sourceNode.level <= targetLevel) {
            return yield* this.syncNode(sourceNode);
        }
        const children = await this.source.getChildren(sourceNode.level, sourceNode.key);
        if (targetLevel === 0 && sourceNode.level === 1) {
            const keys = [];
            for (const { level, key } of children) {
                if (key === null) {
                    continue;
                }
                assert(level === 0, "unexpected leaf level");
                keys.push(key);
            }
            try {
                yield keys;
            }
            catch (err) {
                this.log.error("failed to process batch: %O", err);
            }
        }
        else {
            for (const sourceChild of children) {
                yield* this.syncRoot(sourceChild, targetLevel);
            }
        }
    }
    async *syncNode(sourceNode) {
        const targetNode = this.txn.getNode(sourceNode.level, sourceNode.key);
        if (targetNode !== null) {
            if (targetNode.level === sourceNode.level && equals(targetNode.hash, sourceNode.hash)) {
                return;
            }
        }
        const children = await this.source.getChildren(sourceNode.level, sourceNode.key);
        if (sourceNode.level > 1) {
            for (const sourceChild of children) {
                yield* this.syncNode(sourceChild);
            }
        }
        else if (sourceNode.level === 1) {
            const keys = [];
            for (const { level, key, hash } of children) {
                if (key === null) {
                    continue;
                }
                assert(level === 0, "unexpected leaf level");
                const leaf = this.txn.getNode(0, key);
                if (leaf === null) {
                    keys.push(key);
                }
                else if (equals(hash, leaf.hash)) {
                    continue;
                }
                else {
                    this.log.error("conflict at key %s", hex(key));
                    this.log.error("- target hash: %s", hex(leaf.hash));
                    this.log.error("+ source hash: %s", hex(hash));
                    throw new ConflictError({ level, key, hash }, leaf);
                }
            }
            try {
                yield keys;
            }
            catch (err) {
                this.log.error("failed to process batch: %O", err);
                throw err;
            }
        }
    }
}
